/**
 * Copyright 2008 Peach Jean Solutions
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.xj4;

import org.xj4.spi.MethodWrapperSource;
import org.xj4.spi.ClassWrapperSource;
import org.xj4.spi.TestWrapper;
import org.xj4.spi.MethodHandler;
import org.xj4.spi.MethodHandlerFactory;
import org.xj4.spi.XJ4ExtensionProvider;
import org.xj4.spi.TestClassValidator;

import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.io.IOException;
import java.lang.annotation.Annotation;

import org.junit.runner.notification.RunNotifier;
import org.junit.runners.BlockJUnit4ClassRunner;
import org.junit.runners.model.InitializationError;
import org.junit.runners.model.FrameworkMethod;
import org.junit.runners.model.Statement;
import org.junit.Test;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * This is a JUnit {@link org.junit.runner.Runner} that allows extending the basic functionality of JUnit.  Providers
 * that extend {@link org.xj4.spi.XJ4ExtensionProvider} are provided the opportunity to hook into the test lifecycle at
 * certain places.
 * 
 * @author jbunting
 *
 */
public class XJ4Runner extends BlockJUnit4ClassRunner {

  private static final Log logger = LogFactory.getLog(XJ4Runner.class);

  private XJ4ExtensionProvider extensionProvider;

  private boolean initialized = false;
  private List<MethodHandler> methodHandlers = new ArrayList<MethodHandler>();
  private XJ4TestClass testClass;

  /**
   * Creates a new runner for testing the provided {@code klass}.
   *
   * @param klass the class to test
   * @throws InitializationError if {@code klass} cannot be tested
   */
  public XJ4Runner(Class<?> klass) throws InitializationError {
    this(klass, null);
  }

  /**
   * Creates a new runner for testing the provided {@code klass} with the provided extension provider.
   * @param klass the test class
   * @param extensionProvider the extension provider for this runner
   * @throws InitializationError
   */
  public XJ4Runner(Class<?> klass, XJ4ExtensionProvider extensionProvider) throws InitializationError {
    super(klass);
    this.testClass = new XJ4TestClass(klass);
    try {
      this.extensionProvider = extensionProvider != null ? extensionProvider : new ServiceExtensionProvider();
    } catch (IOException e) {
      throw new InitializationError(Arrays.<Throwable>asList(e));
    }
    for(MethodHandlerFactory factory: this.extensionProvider.getMethodHandlerFactories()) {
      final MethodHandler methodHandler = factory.create(this.getXJ4TestClass());
      if(methodHandler != null) {
        methodHandlers.add(methodHandler);
      }
    }
    initialized = true;
    validate();
  }

  /**
   * Modifies JUnit's test method list by allowing MethodHandlers the opportunity to handle each detected test method.
   * The first MethodHandler willing to handle a method is given that privilege, and if none are willing, then the
   * default JUnit mechanism is used.
   * @return the list of test methods
   */
  @Override
  protected List<FrameworkMethod> computeTestMethods() {
    List<FrameworkMethod> original = super.computeTestMethods();
    List<FrameworkMethod> adjusted = new ArrayList<FrameworkMethod>();
    for(FrameworkMethod method: original) {
      boolean handled = false;
      for(MethodHandler handler: methodHandlers) {
        if(handler.canHandle(method)) {
          adjusted.addAll(handler.handle(method));
          handled = true;
          break;
        }
      }
      if(!handled) {
        adjusted.add(method);
      }
    }
    return adjusted;
  }

  /**
   * Modifies JUnit statement for a given framework method by combining it with all applicable method wrappers provided
   * by extensions.
   * @param frameworkMethod the framework method
   * @param target the target object
   * @param statement the statement
   * @return the statement wrapped with extension functionality
   */
  @Override
  protected Statement withAfters(FrameworkMethod frameworkMethod, Object target, Statement statement) {
    final List<TestWrapper> wrappers = new ArrayList<TestWrapper>();
    for(MethodWrapperSource source: this.extensionProvider.getMethodWrapperSources()) {
      wrappers.addAll(source.generateWrapper(this.getXJ4TestClass(), frameworkMethod, target));
    }
    return new TestWrapperStatement(super.withAfters(frameworkMethod, target, statement), wrappers);
  }

  /**
   * Modifies JUnit statement for a given run notifier by combining it with all applicable class wrappers provided by
   * extensions.
   * @param runNotifier the run notifier
   * @return the statement wrapped with extension functionality
   */
  @Override
  protected Statement classBlock(RunNotifier runNotifier) {
    List<TestWrapper> wrappers = new ArrayList<TestWrapper>();
    for(ClassWrapperSource source: this.extensionProvider.getClassWrapperSources()) {
      wrappers.addAll(source.generateWrapper(this.getXJ4TestClass()));
    }
    return new TestWrapperStatement(super.classBlock(runNotifier), wrappers);
  }

  /**
   * Provides the test class as an XJ4TestClass - useful for extensions.
   * @return the test class
   */
  public XJ4TestClass getXJ4TestClass() {
    return this.testClass;
  }

  /**
   * Validates test methods.  This needs to override the parent since if the method is handled by an extension, then
   * that handler is responsible for all validation.
   * @param errors
   */
  @Override
  protected void validateTestMethods(List<Throwable> errors) {
    validateMethods(Test.class, errors);
  }

  /**
   * Validates all methods annotated with a specific annotation.  This method first gives method handlers an opportunity
   * to validate the method.  The first method handler that claims to handle it is asked to validate it.  If no handlers
   * handle it, then the default JUnit validation is performed.
   * @param annotation the annotation to filter methods by
   * @param errors validation errors
   */
  protected void validateMethods(Class<? extends Annotation> annotation, List<Throwable> errors) {
    List<FrameworkMethod> methods= getTestClass().getAnnotatedMethods(annotation);

    for (FrameworkMethod method : methods) {
      boolean handled = false;
      for(MethodHandler handler: methodHandlers) {
        if(handler.canHandle(method)) {
          handler.validate(method, errors);
          handled = true;
          break;
        }
      }
      if(!handled) {
        method.validatePublicVoidNoArg(false, errors);
      }
    }
  }


  /**
   * Performs validation of the test class.  Note that this re-calls {@link #collectInitializationErrors} since we have
   * made it ineffectual when the super class calls it.
   * @throws InitializationError
   */
  protected void validate() throws InitializationError {
    List<Throwable> errors= new ArrayList<Throwable>();
    // calling the super method here, since in this class, we override it to do nothing
    collectInitializationErrors(errors);
    if (!errors.isEmpty())
      throw new InitializationError(errors);
  }

  /**
   * This method is the initialization validator.  It does two things.  First, it refuses to run validate when the
   * super class attempts to call it.  This is due to the fact that the super class calls this method prior to this
   * class having the opportunity to initialize.  Once this class is initialized, this method is called again.  It then
   * invokes the super method, and then gives each class validator the opportunity to validate the test class.
   * @param errors validation errors
   */
  @Override
  protected void collectInitializationErrors(List<Throwable> errors) {
    // if not initialized, do nothing, the parent calls this before we have a chance to setup our class,
    // causing all sorts of issues
    if(initialized) {
      super.collectInitializationErrors(errors);
      for(TestClassValidator validator : extensionProvider.getTestClassValidators()) {
        validator.validate(this.getXJ4TestClass(), errors);
      }
    }
  }
}
