/**
 * Copyright 2008 Peach Jean Solutions
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.xj4;

import java.lang.reflect.Field;
import java.lang.annotation.Annotation;
import java.util.List;
import java.util.ArrayList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.junit.runners.model.TestClass;

/**
 * Extension of JUnit's {@link TestClass} that provides extension access to certain annotated fields.
 *
 * @author jbunting
 */
public class XJ4TestClass extends TestClass {

  private static final Log logger = LogFactory.getLog(XJ4TestClass.class);

  private Class<?> javaClass;

  /**
   * Instantiates a new test class object.
   * @param javaClass the test class
   */
  public XJ4TestClass(Class<?> javaClass) {
    super(javaClass);
    this.javaClass = javaClass;
  }

  /**
   * Returns a list of fields in the test class that are annotated with the provided annotation class
   * @param annotationClass the annotation class to look for
   * @return the list of fields
   */
  public List<Field> getAnnotatedFields(Class<? extends Annotation> annotationClass) {
    List<Field> results= new ArrayList<Field>();
    return getAnnotatedFields(annotationClass, results);
  }

  /**
   * Populates <code>results</code with a list of fields in the test class that are annotated with the provided
   * annotation class.
   * @param annotationClass the annotation class to look for
   * @param results the list to populate with the results
   * @return results
   */
  public List<Field> getAnnotatedFields(Class<? extends Annotation> annotationClass, List<Field> results) {
    logger.trace("Retrieving fields annotated with " + annotationClass.getName());
    for (Class<?> eachClass : getSuperClasses(javaClass)) {
      Field[] fields= eachClass.getDeclaredFields();
      for (Field eachField : fields) {
        if(logger.isTraceEnabled()) {
          logger.trace("Checking field: " + eachField.getDeclaringClass().getName() + " : " + eachField.getName());
          Annotation[] annotations = eachField.getAnnotations();
          for(Annotation annotation: annotations) {
            logger.trace("Annotation: " + annotation.getClass().getName());
          }
        }
        if (eachField.isAnnotationPresent(annotationClass) && ! isShadowed(eachField, results)) {
          logger.trace("Found one!");
          results.add(eachField);
        }
      }
    }
    return results;
  }

  private boolean isShadowed(Field field, List<Field> results) {
		for (Field each : results) {
			if (isShadowed(field, each))
				return true;
		}
		return false;
	}

	private boolean isShadowed(Field current, Field previous) {
		return previous.getName().equals(current.getName());
	}

  private List<Class<?>> getSuperClasses(Class< ?> testClass) {
    ArrayList<Class<?>> results= new ArrayList<Class<?>>();
    Class<?> current= testClass;
    while (current != null) {
      results.add(current);
      current= current.getSuperclass();
    }
    return results;
  }
}
